import copy
from dataclasses import dataclass
import os
import queue
import re
import threading
import time
import urllib.parse as urlparse
from typing import Optional

from bs4 import BeautifulSoup
import requests

from dokuWikiDumper.utils.util import smkdirs, uopen
from dokuWikiDumper.utils.util import print_with_lock as print
from dokuWikiDumper.utils.config import runtime_config

sub_thread_error = None

@dataclass
class DumpMediaParams:
    dump_dir: str
    title: str
    title_index: int
    base_url: str
    session: requests.Session
    fetch_url: str


def getFiles(url, ns: str = '',  dumpDir: str = '', session: requests.Session=None):
    """ Return a list of media filenames of a wiki """

    if dumpDir and os.path.exists(dumpDir + '/dumpMeta/files.txt'):
        with uopen(dumpDir + '/dumpMeta/files.txt', 'r') as f:
            files = f.read().splitlines()
            if files[-1] == '--END--':
                print('Loaded %d files from %s' %
                      (len(files) - 1, dumpDir + '/dumpMeta/files.txt'))
                return files[:-1]  # remove '--END--'

    files = set()
    ajax = urlparse.urljoin(url, 'lib/exe/ajax.php')
    medialist = BeautifulSoup(
        session.post(ajax, {
            'call': 'medialist',
            'ns': ns,
            'do': 'media'
        }).text, runtime_config.html_parser)
    medians = BeautifulSoup(
        session.post(ajax, {
            'call': 'medians',
            'ns': ns,
            'do': 'media'
        }).text, runtime_config.html_parser)
    imagelinks = medialist.find_all(
        'a',
        href=lambda x: x and re.findall(
            '[?&](media|image)=',
            x))
    for a in imagelinks:
        query = urlparse.parse_qs(urlparse.urlparse(a['href']).query)
        key = 'media' if 'media' in query else 'image'
        files.add(query[key][0])
    files = list(files)
    namespacelinks = medians.find_all('a', {'class': 'idx_dir', 'href': True})
    for a in namespacelinks:
        query = urlparse.parse_qs(urlparse.urlparse(a['href']).query)
        files += getFiles(url, query['ns'][0], session=session)
    print('Found %d files in namespace %s' % (len(files), ns or '(all)'))

    if dumpDir:
        smkdirs(dumpDir + '/dumpMeta')
        with uopen(dumpDir + '/dumpMeta/files.txt', 'w') as f:
            f.write('\n'.join(files))
            f.write('\n--END--\n')

    return files


def dump_media(*, base_url: str, dumpDir: str, session: requests.Session, threads: int = 1, ignore_errors: bool = False):

    smkdirs(dumpDir + '/media')

    fetch = urlparse.urljoin(base_url, 'lib/exe/fetch.php')
    # media_repo = urlparse.urljoin(base_url, '_media')

    files = getFiles(base_url, dumpDir=dumpDir, session=session)

    tasks_queue: queue.Queue[Optional[DumpMediaParams]] = queue.Queue(maxsize=threads)

    workers: list[threading.Thread] = []
    # spawn workers
    for i in range(threads):
        t = threading.Thread(name=f"worker-{i}", target=dump_media_worker, args=(tasks_queue, ignore_errors))
        t.daemon = True
        t.start()
        workers.append(t)

    task_templ = DumpMediaParams(dump_dir=dumpDir, base_url=base_url, session=session, fetch_url=fetch,
                                title_index=-999, title="dokuwikidumper_placehold")

    for index, title in enumerate(files):
        if sub_thread_error:
            raise sub_thread_error

        task = copy.copy(task_templ)
        task.title_index = index
        task.title = title
        tasks_queue.put(task)
        print('Media: (%d/%d): [[%s]] ...' % (index+1, len(files), title))

    for _ in range(threads):
        tasks_queue.put(None)

    tasks_queue.join()

    for w in workers:
        w.join()
        print('worker %s finished' % w.name)

    if sub_thread_error:
        raise sub_thread_error


def dump_media_worker(tasks_queue: queue.Queue[Optional[DumpMediaParams]], ignore_errors: bool):
    global sub_thread_error
    while task := tasks_queue.get():
        try:
            download_media_file(task)
        except Exception as e:
            if not ignore_errors:
                sub_thread_error = e
                raise e
            else:
                print('[',task.title_index,'] Error in sub thread: (', e, ') ignored')
        finally:
            tasks_queue.task_done()
    tasks_queue.task_done()


def download_media_file(task: DumpMediaParams):
    child_path = task.title.replace(':', '/')
    child_path = child_path.lstrip('/')
    child_path = '/'.join(child_path.split('/')[:-1])
    smkdirs(task.dump_dir + '/media/' + child_path)
    file = task.dump_dir + '/media/' + task.title.replace(':', '/')
    local_size = -1
    if os.path.exists(file):
        local_size = os.path.getsize(file)
    with task.session.get(task.fetch_url, params={'media': task.title},
                    stream=True, headers={'Referer': task.base_url}
                    ) as r:
        r.raise_for_status()

        if local_size == -1:  # file does not exist
            to_download = True
        else:
            remote_size = int(r.headers.get('Content-Length', -2))
            if local_size == remote_size:  # file exists and is complete
                print('[%d] File [[%s]] exists (%d bytes)' % (task.title_index+1, task.title, local_size))
                to_download = False
            elif remote_size == -2:
                print('[%d] File [[%s]] cannot get remote size ("Content-Length" missing), ' % (task.title_index+1, task.title) +
                      'will re-download anyway')
                to_download = True
            else:
                to_download = True  # file exists but is incomplete

        if to_download:
            with open(file, 'wb') as f:
                for chunk in r.iter_content(chunk_size=8192):
                    f.write(chunk)
                print('[%d] File [[%s]] Done' % (task.title_index+1, task.title))
        else:
            r.close()

        # modify mtime based on Last-Modified header
        last_modified = r.headers.get('Last-Modified', None)
        if last_modified:
            mtime = time.mktime(time.strptime(
                last_modified, '%a, %d %b %Y %H:%M:%S %Z'))
            atime = os.stat(file).st_atime
            # atime is not modified
            os.utime(file, times=(atime, mtime))



import copy
from dataclasses import dataclass
import os
import queue
import threading
from typing import Optional

import requests
from dokuWikiDumper.dump.content.revisions import get_revisions
from dokuWikiDumper.dump.content.titles import load_get_save_titles
from dokuWikiDumper.utils.util import smkdirs
from dokuWikiDumper.utils.util import print_with_lock as print

from dokuWikiDumper.exceptions import DispositionHeaderMissingError

PDF_DIR = 'pdf/'
PDF_PAGR_DIR = PDF_DIR + 'pages/'
PDF_OLDPAGE_DIR = PDF_DIR + 'attic/'

sub_thread_error = None


@dataclass
class DumpPDFParams:
    dump_dir: str
    title: str
    title_index: int
    doku_url: str
    session: requests.Session
    current_only: bool

def dump_PDF(doku_url, dump_dir,
                  session: requests.Session, threads: int = 1,
                  ignore_errors: bool = False, current_only: bool = False):
    titles = load_get_save_titles(dump_dir=dump_dir, url=doku_url, session=session)
    
    if not len(titles):
        print('Empty wiki')
        return False
    
    tasks_queue: queue.Queue[Optional[DumpPDFParams]] = queue.Queue(maxsize=threads)

    workers: list[threading.Thread] = []
    # spawn workers
    for i in range(threads):
        t = threading.Thread(name=f"worker-{i}", target=dump_pdf_worker, args=(tasks_queue, ignore_errors))
        t.daemon = True
        t.start()
        workers.append(t)

    task_templ = DumpPDFParams(dump_dir=dump_dir, doku_url=doku_url, session=session, current_only=current_only,
                              title_index=-999, title="dokuwikidumper_placehold")

    for index, title in enumerate(titles):
        if sub_thread_error:
            raise sub_thread_error

        task = copy.copy(task_templ)
        task.title_index = index
        task.title = title
        tasks_queue.put(task)
        print('PDF: (%d/%d): [[%s]] ...' % (index+1, len(titles), title))

    for _ in range(threads):
        tasks_queue.put(None)

    tasks_queue.join()

    for w in workers:
        w.join()
        print('worker %s finished' % w.name)

    if sub_thread_error:
        raise sub_thread_error


def dump_pdf_worker(tasks_queue: queue.Queue[Optional[DumpPDFParams]], ignore_errors: bool):
    global sub_thread_error
    while task := tasks_queue.get():
        try:
            dump_pdf_page(task)
        except Exception as e:
            if not ignore_errors:
                sub_thread_error = e
                raise e
            else:
                print('[',task.title_index+1,'] Error in sub thread: (', e, ') ignored')
        finally:
            tasks_queue.task_done()
    tasks_queue.task_done()


def dump_pdf_page(task: DumpPDFParams):
    msg_header = '['+str(task.title_index + 1)+']: '

    file = task.dump_dir + '/' + PDF_PAGR_DIR + task.title.replace(':', '/') + '.pdf'
    local_size = -1
    if os.path.isfile(file):
        local_size = os.path.getsize(file)
    with task.session.get(task.doku_url, params={'do': 'export_pdf', 'id': task.title}, stream=True) as r:
        r.raise_for_status()
        if 'Content-Disposition' not in r.headers:
            raise DispositionHeaderMissingError(r)
        remote_size = r.headers.get('Content-Length', -2)

        if local_size == remote_size:
            print(msg_header, '[[%s]]' % task.title, 'already exists')
        else:
            child_path = task.title.replace(':', '/')
            child_dir = os.path.dirname(child_path)
            smkdirs(task.dump_dir, PDF_PAGR_DIR, child_dir)
            with open(file, 'wb') as f:
                for chunk in r.iter_content(chunk_size=8192):
                    f.write(chunk)
            print(msg_header, '[[%s]]' % task.title, 'saved')

    if task.current_only:
        return True

    revs = get_revisions(doku_url=task.doku_url, session=task.session, title=task.title, msg_header=msg_header)

    for rev in revs[1:]:
        if 'id' in rev and rev['id']:
            try:
                r = task.session.get(task.doku_url, params={'do': 'export_pdf', 'id': task.title, 'rev': rev['id']})
                r.raise_for_status()
                content = r.content
                smkdirs(task.dump_dir, PDF_OLDPAGE_DIR, child_dir)
                old_pdf_path = task.dump_dir + '/' + PDF_OLDPAGE_DIR + child_path + '.' + rev['id'] + '.pdf'

                with open(old_pdf_path, 'bw') as f:
                    f.write(content)
                print(msg_header, '    Revision %s of [[%s]] saved.' % (rev['id'], task.title))
            except requests.HTTPError as e:
                print(msg_header, '    Revision %s of [[%s]] failed: %s' % (rev['id'], task.title, e))